// code by jph

// rotationally symmetric around z:
//    glutSolidCone
//    glutSolidDisc
//    glutSolidDiscbw

// always use RGBA (color4) instead of (color3)

#include <windows.h>
#include <glut.h>

#include "_console.hpp"

#define               KEY_ESC             27
#define               KEY_DL              127
#define               KEY_SPECIAL         256
#define               KEY_AL              KEY_SPECIAL+100
#define               KEY_AR              KEY_SPECIAL+102
#define               KEY_AU              KEY_SPECIAL+101
#define               KEY_AD              KEY_SPECIAL+103
#define               KEY_PU              KEY_SPECIAL+104
#define               KEY_PD              KEY_SPECIAL+105
#define               KEY_P1              KEY_SPECIAL+106
#define               KEY_EN              KEY_SPECIAL+107
#define               KEY_IN              KEY_SPECIAL+108

tensor<int>           opengl_size=tensor<int>(1,2).as(640,480);
tensor<int>           opengl_youtube=tensor<int>(1,2).as(450,360);
tensor<int>           opengl_position=tensor<int>(1,2).as(50,50);
tensor<double>        opengl_project(2,4,4);
tensor<double>        opengl_frustum=tensor<double>(1,2).as(0.1,3000.0);
tensor<char>          opengl_title=tensor<char>::string("");
tensor<int>           keyboard(1,512);
tensor<int>           keyboard_down(1,512);
time                  opengl_time;
time                  opengl_lnch;
time                  opengl_rate;

tensor<int>           mouse(1,2);
tensor<double>        mouse_drag(1,4);
linked<tensor<char>>  opengl_bulletin;
int                   opengl_bulletin_show=1;
int                   opengl_polygonmode=1;
int                   opengl_frame=0;
int                   opengl_shot=0;
int                   opengl_shot_num=0;            // used to enumerate the files:
int                   opengl_shot_sub=0;            // opengl_num-sub.rgba
int                   opengl_recording=0;
int                   opengl_recording_numel=40;    // total shots taken
int                   opengl_recording_frame=20;    // take shot every x frames
int                   opengl_recording_count=0;

tensor<char>          opengl_font("data/white11x07.font"); // courier 7x11 bold

tensor<double>        opengl_color_black   =tensor<double>(1,4).as( 0.0, 0.0, 0.0, 1.0);
tensor<double>        opengl_color_white   =tensor<double>(1,4).as( 1.0, 1.0, 1.0, 1.0);
tensor<double>        opengl_color_red     =tensor<double>(1,4).as( 1.0, 0.0, 0.0, 1.0);
tensor<double>        opengl_color_green   =tensor<double>(1,4).as( 0.0, 1.0, 0.0, 1.0);
tensor<double>        opengl_color_blue    =tensor<double>(1,4).as( 0.0, 0.0, 1.0, 1.0);
tensor<double>        opengl_color_cyan    =tensor<double>(1,4).as( 0.0, 1.0, 1.0, 1.0);
tensor<double>        opengl_color_magenta =tensor<double>(1,4).as( 1.0, 0.0, 1.0, 1.0);
tensor<double>        opengl_color_yellow  =tensor<double>(1,4).as( 1.0, 1.0, 0.0, 1.0);

void display(void); // to be implemented by cpp that includes opengl.hpp

void opengl_display(void) {
  int c0,c1;
  int pix,piy;
  int inc;
  tensor<char> str;
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  opengl_rate=time::now()-opengl_time-opengl_lnch;
  opengl_time+=opengl_rate.s();
  display();

  glLoadIdentity();
  opengl_bulletin.add(tensor<char>::string("rate ")&tensor<double>::string( 1/opengl_rate.s() ));
  //opengl_bulletin.add(tensor<char>::string( opengl_time.s() ));  

  if (opengl_bulletin_show) {
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
      glLoadIdentity();
      gluOrtho2D(0,opengl_size(0),0,opengl_size(1));    
      piy=opengl_size(1);
      inc=opengl_font.cumsize().last();
      while (opengl_bulletin.token(str)) {
        pix=0;
        piy-=opengl_font.size[1];
        for (c0=0;c0<str.numel;c0++) {
          glRasterPos2i(pix,piy);
          switch ((int)str(c0)) {
            case 10:
              pix=0;
              piy-=opengl_font.size[1];
              break;
            default:
              pix+=opengl_font.size[2];
              c1=str(c0);
              if (32<c1&&c1<128&&!keyboard('u'))
                glDrawPixels(opengl_font.size[2],opengl_font.size[1],GL_RGBA,GL_UNSIGNED_BYTE,opengl_font^(c1-32)*inc);
          }
        }
        opengl_bulletin.remove();
      }
    glPopMatrix();
  }
  while (opengl_bulletin.token(str))
    opengl_bulletin.remove();

  if (keyboard('g')) {          //opengl control mode
    opengl_bulletin.add(tensor<char>::string( "opengl\n (b)ulletin\n (s)hot\n (r)ecording\n (w)ireframe" ));
    opengl_bulletin_show^=keyboard_down('b');
    opengl_shot|=keyboard_down('s');
    opengl_recording|=keyboard_down('r');
    if (keyboard_down('w')) {   //wireframe
      opengl_polygonmode^=1;
      if (opengl_polygonmode)
        glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
      else
        glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);      
    }
  }
  
  if (opengl_shot_sub>=opengl_recording_numel) {
    opengl_recording=false;
    opengl_shot_num++;
    opengl_shot_sub=0;
  }
  if (opengl_recording)
    opengl_shot=!(opengl_recording_count++%opengl_recording_frame);
  else
    opengl_recording_count=0;
  
  if (opengl_shot) {
    tensor<unsigned char> rgba(3,4,opengl_size(0),opengl_size(1));
    glReadPixels(0,0,opengl_size(0),opengl_size(1),GL_RGBA,GL_UNSIGNED_BYTE,rgba.data);    
    char fnc[60];
    sprintf(fnc,"out/opengl_%02i-%04i.bmp",opengl_shot_num,opengl_shot_sub);
    format::bmp(fnc,rgba);
    ++opengl_shot_sub;
    opengl_shot=0;    
  }
  keyboard_down=0;
  mouse_drag=0;
  ++opengl_frame;
  
  glutSwapBuffers();
}

void opengl_idle(void) {
  glutPostRedisplay();
}

void opengl_keyboard(unsigned char key,int pix,int piy) {  
  //int mod=glutGetModifiers();
  keyboard(key)=1;
  keyboard_down(key)=1;
}

void opengl_keyboardUp(unsigned char key,int pix,int piy) {  
  if (key==KEY_ESC)
    exit(0);
  keyboard(key)=0;
}

void opengl_special(int key,int pix,int piy) {  
  keyboard(KEY_SPECIAL+key)=1;
  keyboard_down(KEY_SPECIAL+key)=1;
}

void opengl_specialUp(int key,int x,int y) {
  keyboard(KEY_SPECIAL+key)=0;
}

void opengl_mouse(int b,int s,int x,int y) {
  mouse.as(x,y);
}

void opengl_motion(int x,int y) {
  mouse_drag.as(
    (x-mouse(0))/(double)opengl_size(0),
    (mouse(1)-y)/(double)opengl_size(1),
     0.0,
     0.0);
  mouse.as(x,y);
}

void opengl_passiveMotion(int x,int y) {
  mouse.as(x,y);
}

void opengl_entry(int entry) {
#ifndef NDEBUG
  //if (!entry)
  //  exit(0);
#endif
}

void opengl_reshape(int w,int h) { // ok (also called at launch)
  opengl_size.as(w,h);
  glViewport(0,0,w,h);
  glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(55,w/(float)h,opengl_frustum(0),opengl_frustum(1));
    glGetDoublev(GL_PROJECTION_MATRIX,opengl_project.data);
  opengl_display();
}

void opengl_main(int argc,char** argv) {
  int c1;
  glutInit(&argc,argv);
  glutInitDisplayMode(GLUT_DOUBLE|GLUT_DEPTH|GLUT_RGBA);
#ifdef OPENGL_FULLSCREEN
  glutGameModeString(OPENGL_FULLSCREEN);
  if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE))
    glutEnterGameMode();
  else {
    printf("OPENGL_FULLSCREEN error!\n");
    exit(0);
  }
#else
  glutInitWindowSize(opengl_size(0),opengl_size(1));
  glutInitWindowPosition(opengl_position(0),opengl_position(1));
  glutCreateWindow((opengl_title&0).data);
#endif
  glutDisplayFunc(opengl_display);                //callbacks
  glutIdleFunc(opengl_idle);
  glutKeyboardFunc(opengl_keyboard);
  glutKeyboardUpFunc(opengl_keyboardUp);
  glutSpecialFunc(opengl_special);
  glutSpecialUpFunc(opengl_specialUp);
  glutMouseFunc(opengl_mouse);
  glutMotionFunc(opengl_motion);
  glutPassiveMotionFunc(opengl_passiveMotion);
  glutEntryFunc(opengl_entry);
  glutReshapeFunc(opengl_reshape);
  glutIgnoreKeyRepeat(true);
  
  glShadeModel(GL_SMOOTH);              //GL_FLAT
  glClearDepth(1.0f); // TODO what does this do ?!?
  glEnable(GL_DEPTH_TEST);              //depthtest
  glDepthFunc(GL_LESS);
  glPixelStorei(GL_UNPACK_ALIGNMENT,1); //texture
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA); 
  glEnable(GL_BLEND);

  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_LIGHT_MODEL_TWO_SIDE,GL_AMBIENT_AND_DIFFUSE);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,1); // specular light: with respect to 0-> z axis, 1 -> observer
  glEnable(GL_LIGHTING);
  glClearColor(.2,.2,.2,0);

  float fog_color[4]={.15,.15,.15,1};
  
  //glFogi(GL_FOG_MODE, GL_EXP); // GL_LINEAR, GL_EXP2 Fog Mode
  //glFogfv(GL_FOG_COLOR, fog_color); // Set Fog Color
  //glFogf(GL_FOG_DENSITY, 0.35f); // How Dense Will The Fog Be
  //glHint(GL_FOG_HINT, GL_NICEST); // Fog Hint Value
  //glFogf(GL_FOG_START, 1.0f); // Fog Start Depth
  //glFogf(GL_FOG_END, 5.0f); // Fog End Depth
  //glEnable(GL_FOG); // Enables GL_FOG  
  
  
  //glEnable(GL_LINE_SMOOTH); does not look good if drawing order is not ordered
  glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(1,1);
  glEnable(GL_CULL_FACE);
  glEnable(GL_POLYGON_OFFSET_LINE);
  glPolygonOffset(1,100);
  opengl_lnch=time::now();
#ifdef NDEBUG
  opengl_bulletin_show=0;
#endif
}

void opengl_grid() {
  glNormal3d( 0.0, 0.0, 1.0);
  double u=-0.1f; // shift down along z a tiny bit
  glLineWidth(2);
  glColor4d( 0.7, 0.4, 0.0, 0.8);
  glBegin(GL_LINES);
    glVertex3d(-20.0,  0.0,  u  ); glVertex3d( 20.0,  0.0,  u  );
    glVertex3d(  0.0,-20.0,  u  ); glVertex3d(  0.0, 20.0,  u  );
    glVertex3d(  0.0,  0.0,-20.0); glVertex3d(  0.0,  0.0, 20.0);
  glEnd();
  glColor4d( 0.3, 0.7, 0.3, 0.8);
  glLineWidth(1);
  glBegin(GL_LINES);
    for (int c0=-10;c0<11;++c0) {
      glVertex3d(-10.0, c0  ,  u); glVertex3d( 10.0, c0  ,  u);
      glVertex3d( c0  ,-10.0,  u); glVertex3d( c0  , 10.0,  u);
    }
  glEnd();
  u=50;
  double h=-50;
  glColor4d( 1.0, 1.0, 1.0, 1.0);
  glBegin(GL_QUADS);
    glVertex3d(-u, u, h);
    glVertex3d(-u,-u, h);
    glVertex3d( u,-u, h);
    glVertex3d( u, u, h);
  glEnd();
}

void opengl_grid_xy() {
  glPushAttrib(GL_LIGHTING_BIT);      
  glPushAttrib(GL_DEPTH_BUFFER_BIT);      
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    //glHint(GL_LINE_SMOOTH_HINT,GL_DONT_CARE); // GL_DONT_CARE
    glColor4d( 0.3, 0.3, 0.3, 1.0);
    glLineWidth(1);
    glBegin(GL_LINES);
      for (int c0=-10;c0<=10;++c0) {
        glVertex3f(-10, c0, -5);
        glVertex3f( 10, c0, -5);
        glVertex3f( c0,-10, -5);
        glVertex3f( c0, 10, -5);
      }
    glEnd();
    //glHint(GL_LINE_SMOOTH_HINT,GL_NICEST); // GL_DONT_CARE    
  glPopAttrib();
  glPopAttrib();
}

tensor<double> glD(const tensor<double>& myTensor) { // ok
  assert(3<=myTensor.numel);
  return tensor<double>(2,4,4).as(
    0.,myTensor.data[2],-myTensor.data[1],0.,
    -myTensor.data[2],0.,myTensor.data[0],0.,
    myTensor.data[1],-myTensor.data[0],0.,0.,
    0.,0.,0.,0.);
}

tensor<double> glT(const tensor<double>& myTensor) { // ok
  return tensor<double>(2,4,4).as(
    1.,0.,0.,0.,
    0.,1.,0.,0.,
    0.,0.,1.,0.,
    myTensor.data[0],myTensor.data[1],myTensor.data[2],1.);
}

tensor<double> glI(const tensor<double>& myTensor) { // ok
  return tensor<double>(2,4,4).as(
     myTensor.data[ 0],myTensor.data[ 4],myTensor.data[ 8],0.,
     myTensor.data[ 1],myTensor.data[ 5],myTensor.data[ 9],0.,
     myTensor.data[ 2],myTensor.data[ 6],myTensor.data[10],0.,
    -myTensor.data[12]*myTensor.data[ 0]-myTensor.data[13]*myTensor.data[ 1]-myTensor.data[14]*myTensor.data[ 2],
    -myTensor.data[12]*myTensor.data[ 4]-myTensor.data[13]*myTensor.data[ 5]-myTensor.data[14]*myTensor.data[ 6],
    -myTensor.data[12]*myTensor.data[ 8]-myTensor.data[13]*myTensor.data[ 9]-myTensor.data[14]*myTensor.data[10],1.);
}

tensor<double> glN(tensor<double> myTensor) {
  tensor<double> nrm=myTensor.normal(2);
  return tensor<double>(2,4,4).as(
    nrm.data[ 0]*nrm.data[ 0],nrm.data[ 1]*nrm.data[ 0],nrm.data[ 2]*nrm.data[ 0],0.,
    nrm.data[ 0]*nrm.data[ 1],nrm.data[ 1]*nrm.data[ 1],nrm.data[ 2]*nrm.data[ 1],0.,
    nrm.data[ 0]*nrm.data[ 2],nrm.data[ 1]*nrm.data[ 2],nrm.data[ 2]*nrm.data[ 2],0.,
    0.,0.,0.,1.);
}

tensor<double> gl3v(tensor<double> myTensor) {
  return tensor<double>(1,3).as(
    myTensor.data[0],
    myTensor.data[1],
    myTensor.data[2]);
}

tensor<double> gl4v(tensor<double> myTensor) {
  return tensor<double>(1,4).as(
    myTensor.data[0],
    myTensor.data[1],
    myTensor.data[2],
    0.0);
}

tensor<double> gl3m(tensor<double> myTensor) { // ok
  return tensor<double>(2,3,3).as(
    myTensor.data[ 0],myTensor.data[ 1],myTensor.data[ 2], // appears in code as 'transposed'
    myTensor.data[ 4],myTensor.data[ 5],myTensor.data[ 6],    
    myTensor.data[ 8],myTensor.data[ 9],myTensor.data[10]);
}

tensor<double> gl4r(tensor<double> myTensor) { // ok
  return tensor<double>(2,4,4).as(
    myTensor.data[ 0],myTensor.data[ 1],myTensor.data[ 2],0.0, // appears in code as 'transposed'
    myTensor.data[ 4],myTensor.data[ 5],myTensor.data[ 6],0.0,    
    myTensor.data[ 8],myTensor.data[ 9],myTensor.data[10],0.0,
    0.0,0.0,0.0,1.0);
}

tensor<double> glMatrixFromVector(tensor<double> dir,tensor<double> pos) {
  if (TENSOR_EPS<dir.norm(2)) {
    tensor<double> vc0=gl3v(dir);
    vc0=vc0.normal(2);
    int ind;
    vc0.absolute().min_(ind);
    tensor<double> vc1(1,3);
    vc1.data[ind]=1;
    vc1=vc1-vc0*(vc0|vc1);
    vc1=vc1.normal(2);
    tensor<double> ret(2,4,4);
    ret.in( 0,vc1);
    //ret.in( 4,glX(vc0,vc1));
    ret.in( 4,vc0.x(vc1));
    ret.in( 8,vc0);
    ret.in(12,pos.data,3);
    ret.data[15]=1;
    return ret;
  } else
    return tensor<double>::eye(4);
}

// this is probably for RAPID ????
void inline gl3a(double vout[3][3],tensor<double> myTensor) {
  vout[0][0]=myTensor.data[0];
  vout[1][0]=myTensor.data[1];
  vout[2][0]=myTensor.data[2];
  vout[0][1]=myTensor.data[3];
  vout[1][1]=myTensor.data[4];
  vout[2][1]=myTensor.data[5];
  vout[0][2]=myTensor.data[6];
  vout[1][2]=myTensor.data[7];
  vout[2][2]=myTensor.data[8];
};

// =========================================================================
// DRAWING FUNCTIONS

// 2D -----------------------------------

void glutSolidDisc(double rad,int slices) { // ok
  double rab=abs(rad);
  double fac=MATH_2PI/slices;
  glBegin(GL_TRIANGLE_FAN);
    glNormal3d(0,0,math::Sign(rad));
    glVertex3d(0,0,0);
    for (int c0=0;c0<=slices;++c0)      
      glVertex3d(rab*cos(c0*fac),rad*sin(c0*fac),0);      
  glEnd();
}

void glutSolidDiscbw(double rad,int slices,int swp,tensor<double> mas) {
  assert(0<=rad);
  assert(mas.numel==4);
  double fac=MATH_2PI/slices;  
  double lam=0;
  double ang;
  tensor<double> col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, mas(3));
  glBegin(GL_TRIANGLE_FAN);
    glColor4dv(col.data);
    glNormal3d(0,0,math::Sign(rad));
    glVertex3d(0,0,0);
    for (int c0=0;c0<=slices;++c0) {
      if (c0%swp==0)
        lam=1-lam;
      ang=(c0+0.5)*fac;
      glColor4dv((col*lam+mas*(1-lam)).data);
      glVertex3d(rad*cos(ang),rad*sin(ang),0);      
    }
  glEnd();
}

// 3D -----------------------------------

void glutSolidCylinder(double gnd,double hgt,double rad,int slices) {
  double fac=MATH_2PI/slices;
  double cc0,sc0;
  glBegin(GL_QUAD_STRIP);
  for (int c0=0;c0<=slices;++c0) {
    glNormal3d(cc0=cos(c0*fac),sc0=sin(c0*fac),0);
    glVertex3d(rad*cc0,rad*sc0,hgt);
    glVertex3d(rad*cc0,rad*sc0,gnd);
  }
  glEnd();        
}

void glutSolidCylinder(double gnd,double hgt,double rad,double pea,int slices) {
  double fac=MATH_2PI/slices;
  double cc0,sc0;
  glBegin(GL_QUAD_STRIP);
  for (int c0=0;c0<=slices;++c0) {
    glNormal3d(cc0=cos(c0*fac),sc0=sin(c0*fac),0);
    glVertex3d(pea*cc0,pea*sc0,hgt);
    glVertex3d(rad*cc0,rad*sc0,gnd);
  }
  glEnd();        
}

void glutSolidCylinderbw(double gnd,double hgt,double rad,int slices,int swp,tensor<double> mas) {
  assert(mas.numel==4);
  double fac=MATH_2PI/slices;
  double lam=0,cc0,sc0;
  double ang;
  tensor<double> col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, mas(3));
  glBegin(GL_QUAD_STRIP);
    for (int c0=0;c0<=slices;++c0) {
      if (c0%swp==0)
        lam=1-lam;
      ang=(c0+0.5)*fac;
      glNormal3d(cc0=cos(ang),sc0=sin(ang),0);
      glColor4dv((col*lam+mas*(1-lam)).data);
      glVertex3d(rad*cc0,rad*sc0,hgt);
      glVertex3d(rad*cc0,rad*sc0,gnd);
    }
  glEnd();
  glPushMatrix();
    glTranslatef(0,0,hgt);
    glutSolidDiscbw( rad,slices,swp,mas);
  glPopMatrix();
}

void glutSolidSpherebw(double beg,double end,double rad,int slices,int swp,tensor<double> mas) {
  assert(mas.numel==4);
  assert(beg<end);
  assert(0<=rad);
  int cuts=(end-beg)*slices/MATH_2PI;
  double fac=MATH_2PI/slices;
  double mul=(end-beg)/cuts;
  double lam=0;
  tensor<double> col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, mas(3));
  tensor<double> dir(1,3);
  tensor<double> rot=tensor<double>(2,3,3).as(cos(fac),sin(fac),0.0,-sin(fac),cos(fac),0.0,0.0,0.0,1.0);
  for (int c0=0;c0<slices;++c0) {
    if (c0%swp==0)
      lam=1-lam;
    glBegin(GL_QUAD_STRIP);
    for (int c1=0;c1<=cuts;++c1) {
      glColor4dv((col*lam+mas*(1-lam)).data);
      dir.as(cos(c0*fac)*sin(c1*mul+beg),sin(c0*fac)*sin(c1*mul+beg),-cos(c1*mul+beg));
      glNormal3dv(dir.data); glVertex3dv((dir*rad).data);
      dir=rot*dir;
      glNormal3dv(dir.data); glVertex3dv((dir*rad).data);      
    }
    glEnd();
  }
  glPushMatrix();
    glTranslatef(0,0,-rad*cos(end));
    glutSolidDiscbw( rad*sin(end),slices,swp,mas); // TODO increase the radius a bit
  glPopMatrix();
}
