// code by jph


//----------------------------------------------------------------------

void tensor_toTest(void) {
  tensor<double> asd=tensor<double>::rand(2,3,3);
  tensor<double> ret;
  tensor<double> qrt=asd.schur(ret);
  printf("%s\n---\n",asd.toString());
  printf("%s\n---\n",ret.chop().toString());
  printf("%s\n---\n",(qrt*ret*qrt.transpose()).toString());
  printf("double\n%s\n", asd.toString());
  printf("int   \n%s\n",(asd*0x7fffffff).to<int>().toString("%08x "));
  printf("--------------------------\n");
  //asd.to<double>();
}

template<class type>
void tensor_svd_sparse_toTest(void) {
  tensor<tensor<type>*> blk(2,2,2);
  //blk(0,0)=new tensor<type>::rand(2,3,3);
  blk(0,0)=new tensor<type>(tensor<type>::rand(2,3,3));
  blk(1,0)=new tensor<type>(2,3,3);
  blk(0,1)=new tensor<type>(2,3,3);
  blk(1,1)=new tensor<type>(2,3,3);
  tensor<type> MAT(blk);
  tensor<type> w,v;  
  tensor<type> d=MAT.svd(w,v);
  printf("%s\n",d.toString());
  printf("%s\n",v.toString());
  printf("%s\n",w.toString());
  printf("------------------\n");
  printf("%s\n",(d.transpose()*tensor<type>::diag(w)*v.transpose()).toString());
  printf("------------------\n");
  printf("%s\n",MAT.toString());
  printf("------------------\n");
}

template<class type>
void tensor_block_toTest(void) {
  tensor<tensor<type>*> bla(2,4,3);
  for (int c0=0;c0<bla.numel;++c0)
    bla.data[c0]=new tensor<type>(2,1,2);
  tensor<type> BLA(bla);
}

template<class type>
void tensor_string_toTest(void) {
  tensor<type> asd=tensor<type>::index(2,2,3)+32;
  printf("%s\n",asd.toString());
  tensor<char> txt=tensor<type>::string((type)123.2);
  printf("-----> %s\n",txt.toString());
}

void tensor_svd_toTest(void) {
  tensor<double> svdt=tensor<double>::randn(2,11,5);
  for (int c1=0;c1<svdt.size[0];c1++)
    svdt(c1,3)=0;
  printf("%s\n\n",svdt.toString());
  tensor<double> uor;
  tensor<double> dia;
  tensor<double> ort;
  tensor<double>::valid=0;
  printf("valid=%i\n",tensor<double>::valid);
  uor=svdt.svd(dia,ort);
  printf("valid=%i\n",tensor<double>::valid);
  printf("%s\n\n",(uor*uor.transpose()).toString());
  printf("%s\n\n",(ort.transpose()*ort).toString());
  printf("%s\n\n",(ort*ort.transpose()).toString());
  dia=tensor<double>::diag(dia);
  printf("%s\n\n",dia.toString());
  printf("difference=\n%s\n\n",(uor.transpose()*dia*ort.transpose()-svdt).toString());
  //printf("%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
}

template<class type>
void tensor_inverse_toTest() {
  tensor<type> rnd=tensor<type>::rand(2,7,5);
  tensor<type> spl=tensor<type>::eye(5);
  spl(4,3)=spl(3,4)=spl(1,2)=spl(2,1)=1;
  rnd=rnd*spl;
  printf("FOR MATLAB\n%s\n---------\n",rnd.toString(" %14.4e"));
  tensor<type> ide=rnd.inverse()*rnd;  
  printf("OUR PINVERSE:\n%s\n--------------\n",rnd.inverse().toString(" %14.4e"));
  printf("%s\n",ide.toString());
  tensor<type> fgh=tensor<type>::randn(2,7,7);
  tensor<type> dif=fgh.solve(tensor<type>::eye(7))-fgh.inverse();
  printf("inv=pinv :\n%s\n",dif.toString());
}

void tensor_rank_toTest(void) {
  tensor<double> svdt=tensor<double>::randn(2,11,11);
  for (int c1=0;c1<svdt.size[0];c1++) {    
    //svdt(c1,1)=TENSOR_NO_DIVISION;
    //svdt(c1,3)=0;
  }
  printf("rank = %i\n\n",svdt.transpose().rank());
}

void tensor_sort_inv_test() {
  tensor<int> rnd=(tensor<float>::rand(1,15)*100).to<int>();
  tensor<int> ind;
  tensor<int> srt=rnd.tros(ind);
  //tensor<int> srt=rnd.sort(ind);
  printf("%s\n",srt.toString());
}

void tensor_svd_toTest2(void) {
  //tensor<double> svdt=tensor<double>::randn(2,6,6);
  tensor<double> svdt(2,5,5);//=tensor<double>::randn(2,6,6);
  //for (int c1=0;c1<svdt.size[0];c1++) {
  //  svdt(c1,3)=0;
  //  svdt(c1,4)=0;
  //  svdt(c1,5)=0;
  //}
  printf("%s\n\n",svdt.toString());
  tensor<double> uor;
  tensor<double> dia;
  tensor<double> ort;
  tensor<double>::valid=0;
  printf("valid=%i\n",tensor<double>::valid);
  uor=svdt.svd(dia,ort);
  printf("valid=%i\n",tensor<double>::valid);
  printf("%s\n\n",uor.toString());
  printf("%s\n\n",(uor.transpose()*uor).toString());
  printf("%s\n\n",(ort.transpose()*ort).toString());
  printf("%s\n\n",(ort*ort.transpose()).toString());
  dia=tensor<double>::diag(dia);
  printf("%s\n\n",dia.toString());
  printf("difference=\n%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
  //printf("%s\n\n",(uor*dia*ort.transpose()-svdt).toString());
}

void sha_toTest() {
  sha mySha;
  char* stra=new char[16];
  stra[ 0]=0x0;
  stra[ 1]=0x1;
  stra[ 2]=0x2;
  stra[ 3]=0x3;
  stra[ 4]=0x4;
  stra[ 5]=0x5;
  stra[ 6]=0x6;
  stra[ 7]=0x7;
  stra[ 8]=0x8;
  stra[ 9]=0x9;
  stra[10]=0x10;
  stra[11]=0x11;
  stra[12]=0x12;
  stra[13]=0x13;
  stra[14]=0x14;
  stra[15]=0x15;
  mySha.add(stra,16);
  mySha.add(stra,16);
  mySha.add(stra,16);
  mySha.add(stra,16);
  //mySha.add(stra,16);
  //mySha.add(stra,10);
  //mySha.add(stra,10);
  //mySha.add(stra,10);
  printf("%s\n",mySha.toInfo());
  printf("%s\n",mySha.toString());
  //tensor<unsigned int> asd2=mySha.flush();
  tensor<int> asd=mySha.flush();
  printf("\n%s\n3fc78b03 7eb94ea8 d0cc96b2 c1a172b7 3df20a50\n",asd.toString("%08x "));
  //printf("%s\n",mySha.toInfo());
  //printf("%s\n",mySha.toString());
  //printf("%08x\n",asd.sum());
  //mySha.add(stra,3);
  //printf("%s\n",mySha.toInfo());
  //asd=mySha.flush();  
  //printf("%s\n",asd.toString("%08x "));
  //printf("%08x\n",asd.sum());
  //tensor<int> dist(1,13);  
  //for (int c1=0;c1<10000;c1++) {
  //  tensor<double> some=tensor<double>::rand(2,10,4);
  //  //printf("%s\n",some.toInfo());
  //  //printf("%s\n",some.toString());
  //  //printf("%08x\n",sha_hash(some));
  //  //printf("%08x\n",sha::hash<double>(some));
  //  //printf("%i\n",sha::hash<double>(some));
  //  dist.data[sha::hash<double>(some)%13]++;
  //}
  //for (int c1=0;c1<13;c1++) 
  //  printf("%3i %i\n",c1,dist.data[c1]);
}

void random_toTest(void) {
  //random<double>::randomize();
  printf("%s\n",random<double>::toInfo());
  double d1=random<double>::uniform();
  printf("%s\n",random<double>::toInfo());
  double n1,n2;
  for (int c1=0;c1<10;c1++) {
    n1=random<double>::normal();
    printf("%f\n",n1);
  }
}

//----------------------------------------------------------------------

void complex_toTest() {
  complex<double> a(-2,3);
  complex<double> b(5,9);
  printf("%s\n",(a/b).toInfo());
  printf("%s\n",(a*b).toInfo());
}

void node_toTest(void) {
  node<int> myNodeA(0,3);
  node<int> myNodeB(&myNodeA,2);
  printf("%s\n",myNodeA.toInfo());
  printf("%s\n",myNodeB.toInfo());
}

//----------------------------------------------------------------------

void linked_toTest(void) {
  int myInt;
  linked<int> myLinked;
  printf("%s\n",myLinked.toInfo());
  myLinked.add(0);
  printf("%s <- 1 added\n",myLinked.toInfo());
  while (myLinked.token(myInt))
   myLinked.remove();
  printf("%s <- 1 removed \n",myLinked.toInfo());
  myLinked.add(1);
  myLinked.add(2);
  myLinked.add(3);
  myLinked.add(4);
  myLinked.add(5);
  myLinked.add(6);
  printf("%s\n",myLinked.toInfo());
  
  while (myLinked.token(myInt)) {
    printf("token %i\n",myInt);
    if (myInt==3||myInt==1) {
      myLinked.remove();
      myLinked.add(0);      
      myLinked.pre(-1);
    }
  }
  printf("%s\n",myLinked.toInfo());
  while (myLinked.token(myInt))
    printf("token %i\n",myInt);
  //myLinked.token_reset();
  printf("----\n");
  printf("%s\n",myLinked.toInfo());  
  while (myLinked.token(myInt)) {
    printf("token %i\n",myInt);
  }
  tensor<int> myTensor=myLinked.toTensor();
  printf("tensor %s\n",myTensor.toString());
  printf("%s\n",myLinked.toInfo());
  while (myLinked.token(myInt))
    myLinked.remove();    
  printf("%s <- after while remove \n",myLinked.toInfo());  
  myLinked.empty();
  printf("%s\n",myLinked.toInfo());  
  printf("-------\n");
  linked<int>* myStar=new linked<int>();
  printf("%s\n",myStar->toInfo());
  myStar->add(0);
  myStar->add(1);
  printf("%s\n",myStar->toInfo());
  delete myStar;
  //printf("%s\n",myStar->toInfo());  
}

int hash_toTest_int(int myInt) {
  return myInt%HASH_TOTEST;
}

void hash_toTest_int(void) {
  hash<int,float> myHash(HASH_TOTEST,hash_toTest_int);  
  float myFloat=0.23f;
  myHash.add(3,myFloat);
  myFloat=0;
  myHash.add(3,myFloat);
  printf("float %8.3f\n",myFloat);
  myFloat=-10;
  myHash.add(3+5,myFloat);
  myFloat=.3333;
  myHash.add(3+25,myFloat);
  bucket<int,float> buck;
  for (int c0=0;c0<myHash.table.numel;++c0)
    while (myHash.table(c0)->token(buck))
      printf("%i key=%4i data=%8.3f\n",c0,buck.key,buck.data);
  printf("%s\n",myHash.toInfo());
  int status;
  status=myHash.get(3+2,myFloat);
  printf("get 5 %i %6.3f\n",status,myFloat);
  status=myHash.get(3+5,myFloat);
  printf("get 8 %i %6.3f\n",status,myFloat);
  hash_flush(myHash,"myhash.txt");  
}

void hash_toTest_void(void) {
  hash<int,void*> myHash(HASH_TOTEST,hash_toTest_int);  
  void* null=(void*)9999;
  myHash.add(3,null);
  int status;
  status=myHash.get(3,null);
  printf("get 3 %i %08x\n",status,null);
}


int hash_toTest_tensor(tensor<int> myTensor) {
  return myTensor.sum()%HASH_TOTEST;
}

void hash_toTest_tensor(void) {
  hash<tensor<int>,tensor<float>> myHash(HASH_TOTEST,hash_toTest_tensor);
  tensor<int> key(1,3);
  key.as(2,3,4);
  tensor<float> color(1,3);
  color(0)=0.222;
  color(1)=0.111;
  color(2)=0.333;
  myHash.add(key,color);
  key(1)++;
  myHash.add(key,color);
  
  printf("%s\n",myHash.toInfo());
  hash_flush(myHash,"myhash_tensor.txt");
}

