// code by jph

#ifndef TENSOR_INCLUDED
#define TENSOR_INCLUDED

#define TENSOR_EMPTY              ndims=0; \
                                  size=0; \
                                  numel=0; \
                                  data=0;

                                  // consider deleting size !
#define TENSOR_NEW_SIZE           size=new int[ndims]; \
                                  va_list ap; \
                                  va_start(ap,ndims); \
                                  for (int c0=0;c0<ndims;++c0) \
                                    size[c0]=va_arg(ap,int); \
                                  va_end(ap);  

#define TENSOR_NUMEL              numel=1; \
                                  for (int TENSOR_NUMEL_c0=0;TENSOR_NUMEL_c0<ndims;TENSOR_NUMEL_c0++) \
                                    numel*=this->size[TENSOR_NUMEL_c0];

#define TENSOR_FORMAT(val)        if (math::Abs(val)<TENSOR_NO_DIVISION) \
                                    len+=sprintf(str+len,nul); \
                                  else \
                                    len+=sprintf(str+len,fmt,val);

// ideally, eps, no_division, ... should depend on type=double,float,

#define TENSOR_EPS          1E-40
#define TENSOR_NO_DIVISION  1E-9

static char TENSOR_BUFFER[8192];

template<class type>
class tensor {
  public:
    tensor<type>(void); // ok
    inline ~tensor<type>(void); // ok
    tensor<type>(int ndims,int* size); // ok (!also takes care of (1,0)!)
    tensor<type>(int ndims,...); // ok (!can use (1,0,0) instead of (1,0)!)
    tensor<type>(const tensor<type>& myTensor); // ok
    tensor<type>(char* name); // to test
    tensor<type>(tensor<tensor<type>*> block); // ok (!deletes block entries at the end!)
    static tensor<type> eye(int size0); // ok
    static tensor<type> diag(tensor<type> myTensor); // ok (!constructs matrix!)
    static tensor<type> rand(int ndims,...); // ok
    static tensor<type> randn(int ndims,...); // ok
    static tensor<type> index(int ndims,...); //
    static tensor<type> scalar(void); // ok
    static tensor<type> vector(int size0); // ok
    static tensor<type> matrix(int size0,int size1); // ok
    static tensor<char> string(char* copy); // ok
    static tensor<char> string(const type& value);     // ok // todo: use template !    
    //static tensor<char> from(char* copy); // ok
    static tensor<type> file(char* name); // ok
    const tensor<type>& operator=(const tensor<type>& myTensor); //
    type& operator()(int ind0,...); // ok
    tensor<type> operator()(tensor<int> ind0,...); // ok
    type& last(void); // ok
    tensor<type> operator[](int slice); // ok
    const tensor<type>& operator=(const type& data); // ok
    const tensor<type>& as(type data0,...); // for int and doubles ok, but problem with floats
    const tensor<type>& diag(type data0,...); // for int and doubles ok, but problem with floats
    const tensor<type>& in(type* copy); // for int and doubles ok, but problem with floats
    const tensor<type>& in(int ind,type* copy,int len); // ok
    const tensor<type>& in(int ind,tensor<type> copy); //
    template<class comp> tensor<comp> to(void); // ok (!converts tensor<type> to tensor<comp>!)
    const tensor<type>& fscan(FILE* fid); //
    tensor<int> at(int ind); // ok
    tensor<int> cumsize(void);
    tensor<int> find(void);
    tensor<type> diag(void);
    tensor<type> operator*(const tensor<type>& myTensor); // ok
    tensor<type> operator/(const tensor<type>& myTensor); // ok (!outer product!)
    tensor<type> operator%(const tensor<type>& myTensor); // ok (!pointwise product!)
    tensor<type> operator+(const tensor<type>& myTensor); // ok
    tensor<type> operator-(const tensor<type>& myTensor); // ok
    tensor<type> operator&(const tensor<type>& myTensor); // ok
    tensor<type> operator*(type data); // ok
    tensor<type> operator/(type data); // ok
    tensor<type> operator+(type data); // ok
    tensor<type> operator-(type data); // ok
    tensor<int> operator<(type data); // 
    tensor<int> operator>(type data); // 
    //tensor<int> operator==(type data); // 
    int operator==(const tensor<type>& myTensor); // 
    tensor<type> operator&(type app); // ok
    type* operator^(int ind) {return &data[ind];}; // ok
    const tensor<type>& operator+=(const tensor<type>& myTensor); // 
    const tensor<type>& operator-=(const tensor<type>& myTensor); // 
    const tensor<type>& operator*=(type data); // ok
    const tensor<type>& operator/=(type data); // ok
    const tensor<type>& operator+=(type data); // ok
    const tensor<type>& operator-=(type data); // ok
    type operator|(const tensor<type>& myTensor); // ok (!wtf skalarprod??? !)
    tensor<type> x(tensor<type>& myTensor);
    char* toInfo(void); // ok
    char* toString(char* str,char* fmt); // developed    
    char* toString(char* fmt); // ok
    char* toString(void); // ok
    tensor<type> exp(void); // ok (!exp rises fast, beware of overflow!)
    tensor<type> log(void); // ok (!log converges slowly, if at all!)
    tensor<type> solve(tensor<type> rhs); // ok, aber nicht optimiert
    //void reshape(int ndims,...); // TODO: relocate, assert, return *this
    tensor<type> transpose(tensor<int>& sig); // ok
    tensor<type> transpose(void); // ok
    tensor<type> trace(tensor<int>& dims); // ok (!is sum() when dims.numel==1!)
    tensor<type> trace(int dim0,int dim1); // ok
    tensor<type> reverse(void) {
      tensor<type> ret(ndims,size);
      int c1=numel;
      for (int c0=0;c0<numel;++c0)
        ret.data[c0]=data[--c1];
      return ret;
    }
    tensor<type> qr(tensor<type>& ret); // ok (!Householder, q.transpose()*ret==*this!)
    tensor<type> schur(tensor<type>& ret); // often oscilating behaviour
    tensor<type> svd(tensor<type>& w,tensor<type>& v); // ok (!size[1]<=size[0]!, ret'*diag(w)*v'==*this)
    tensor<type> inverse(void); // ok (!gives inverse, or 'pseudo inverse' if rank deficient; uses precision!)
    tensor<type> inverse(int num); //
    tensor<type> null(void); // TODO
    // for least squares fit, use x=A.inverse()*b
    int rank(void); // ok
    type norm(int norm); // ok ({0,1,2})
    tensor<type> normal(int norm);
    inline tensor<type> normal();
    tensor<type> absolute(void);
    tensor<type> sort(tensor<int>& ind); // using quicksort
    tensor<type> tros(tensor<int>& ind); // using quicksort
    tensor<type> apply(type (*func)(type));
    const tensor<type>& apply(type (*func)(void));
    type sum(void);
    type prod(void);
    type min_(int& ind); // rename later!!!
    type max_(int& ind); // rename later!!!    
    tensor<type> chop(void); // rename later!!!    
    tensor<tensor<type>*> split(type sep); // ok
    void store(char *name);
    void flush(char *name);
    void free(void);
//----------------------------------------------------------------------
    int ndims; // number of dimensions
    int *size; // size[0],...,size[ndims-1]
    int numel; // total number of elements
    type *data; // data[0],...,data[numel-1]
//----------------------------------------------------------------------
    static char* format; // after critical operations, this is 1 when ok, or 0 in case of failure
    static int valid; // after critical operations, this is 1 when ok, or 0 in case of failure
    static type precision; // after critical operations, this is 1 when ok, or 0 in case of failure
  private:
    void quicksort(int a,int b,int *ind);
    void quicktros(int a,int b,int *ind);
};

template<class type>
tensor<type>::tensor(void) {
  TENSOR_EMPTY  
}

template<class type>
inline tensor<type>::~tensor(void) {
  if (data) {
    delete[] data;
    delete[] size;
  }
}

template<class type>
tensor<type>::tensor(int ndims,int *size) {
  this->ndims=ndims;
  this->size=new int[ndims];
  if (size)
    for (int c0=0;c0<ndims;++c0)
      this->size[c0]=size[c0];
  else
    this->size[0]=0;
  TENSOR_NUMEL
  data=new type[numel];
  *this=0;
}

template<class type>
tensor<type>::tensor(int ndims,...) {
  this->ndims=ndims;
  TENSOR_NEW_SIZE
  TENSOR_NUMEL
  data=new type[numel];  
  *this=0; // this is a constraint!
}

//template<class type> void tensor<type>::reshape(int ndims,...) {
//  // TODO assert numels stay the same!!
//  this->ndims=ndims;
//  delete size;
//  size=new int[ndims];
//  TENSOR_NEW_SIZE
//  //TENSOR_NUMEL
//  //data=new type[numel];  
//  //*this=0; // this is a constraint!
//}

template<class type>
tensor<type>::tensor(const tensor<type>& copy) {
  int c0;
  if (3<copy.ndims) {
    printf("constr %i\n",copy.ndims);
  }
  ndims=copy.ndims;
  size=new int[ndims];
  for (c0=0;c0<ndims;++c0)
    size[c0]=copy.size[c0];
  numel=copy.numel;
  data=new type[numel];
  for (c0=0;c0<numel;++c0)
    data[c0]=copy.data[c0];
}

template<class type> tensor<type>::tensor(char* name) {
  FILE *fid=fopen(name,"rb");
  if (fid) {
    fread(&ndims,sizeof(int),1,fid); // ndims
    size=new int[ndims];
    fread(size,sizeof(int),ndims,fid); // size
    TENSOR_NUMEL
    data=new type[numel]; // data
    fread(data,sizeof(type),numel,fid);
    fclose(fid);
  } else {
    TENSOR_EMPTY
  }
#ifndef NDEBUG
  printf("%s ->\t%s\n",name,toInfo());
#endif
}

//template<class type> tensor<type>::tensor(FILE *name) {
//  if (fid) {
//    fread(&ndims,sizeof(int),1,fid);
//    size=new int[ndims];
//    fread(size,sizeof(int),ndims,fid);
//    TENSOR_NUMEL
//    data=new type[numel];
//    fread(data,sizeof(type),numel,fid);
//  } else {
//    TENSOR_EMPTY
//  }
//};

template<class type> tensor<type>::tensor(tensor<tensor<type>*> block) {
  int c0,c1;
  int ofs=1;
  tensor<int> size(1,block.ndims);
  for (c0=0;c0<block.ndims;++c0) { // compute return size[]
    for (c1=0;c1<block.size[c0];++c1)
      size(c0)+=block.data[c1*ofs]->size[c0];
    ofs*=block.size[c0];
  }
  ndims=block.ndims;
  this->size=new int[ndims];
  for (c0=0;c0<ndims;++c0)
    this->size[c0]=size(c0);
  TENSOR_NUMEL
  data=new type[numel];
  tensor<int> inj(block.ndims,block.size);
  tensor<int> cum=cumsize();
  tensor<int> cnt(1,block.ndims);
  tensor<int> tmp(1,block.ndims);
  for (c0=0;c0<block.numel-1;++c0) { // compute injection addresses
    tmp.in(block.data[c0]->size);
    for (c1=0;c1<tmp.numel;c1++)
      if ((cnt(c1)+tmp(c1))%size(c1)==0)
        cnt(c1)=0;
      else {
        cnt(c1)+=tmp(c1);
        break;
      }
    inj.data[c0+1]=cnt|cum;
  }
  for (c0=0;c0<block.numel;++c0) { // copy blocks into matrix
    for (c1=0;c1<block.data[c0]->numel;++c1)
      data[inj.data[c0]+(block.data[c0]->at(c1)|cum)]=block.data[c0]->data[c1];
    delete block.data[c0];
  }
}

template<class type> 
const tensor<type>& tensor<type>::operator=(const tensor<type>& myTensor) {
  if (data==myTensor.data)
    return *this;
  if (data) {
    delete[] data;
    delete[] size;
  }
  ndims=myTensor.ndims;
  size=new int[ndims];
  int c0;
  for (c0=0;c0<ndims;++c0)
    size[c0]=myTensor.size[c0];
  TENSOR_NUMEL
  data=new type[numel];
  for (c0=0;c0<numel;++c0)
    data[c0]=myTensor.data[c0];
  return *this;
}

template<class type>
tensor<type> tensor<type>::eye(int size0) {
  tensor<type> ret(2,size0,size0);
  for (int c0=0;c0<ret.numel;c0+=size0+1)
    ret.data[c0]=1;
  return ret;
}

template<class type>
tensor<type> tensor<type>::diag(tensor<type> myTensor) {
  tensor<type> ret(2,myTensor.numel,myTensor.numel);
  for (int c0=0;c0<myTensor.numel;c0++)
    ret(c0,c0)=myTensor.data[c0];
  return ret;
}

template<class type>
tensor<type> tensor<type>::rand(int ndims,...) {
  int* size;
  TENSOR_NEW_SIZE
  tensor<type> myTensor(ndims,size);
  delete[] size;
  for (int c0=0;c0<myTensor.numel;++c0)  
    myTensor.data[c0]=random<type>::uniform();
  return myTensor;
}

template<class type>
tensor<type> tensor<type>::randn(int ndims,...) {
  int* size;
  TENSOR_NEW_SIZE
  tensor<type> myTensor(ndims,size);
  delete[] size;
  for (int c0=0;c0<myTensor.numel;++c0)  
    myTensor.data[c0]=random<type>::normal();
  return myTensor;
}

template<class type>
tensor<type> tensor<type>::index(int ndims,...) {
  int* size;
  TENSOR_NEW_SIZE
  tensor<type> myTensor(ndims,size);
  delete[] size;
  for (int c0=0;c0<myTensor.numel;++c0)
    myTensor.data[c0]=c0;
  return myTensor;
}

template<class type>
tensor<type> tensor<type>::scalar(void) {
  return tensor<type>(0);
}

template<class type>
tensor<type> tensor<type>::vector(int size0) {
  return tensor<type>(1,size0);
}

template<class type>
tensor<type> tensor<type>::matrix(int size0,int size1) {
  return tensor<type>(2,size0,size1);
}

template<class type>
tensor<char> tensor<type>::string(char* copy) {
  return tensor<char>(1,strlen(copy)).in(copy);
}

template<class type>
tensor<char> tensor<type>::string(const type& value) {
  sprintf(TENSOR_BUFFER,tensor<type>::format,value);
  return tensor<char>(1,strlen(TENSOR_BUFFER)).in(TENSOR_BUFFER);
}

template<class type>
tensor<type> tensor<type>::file(char* name) {
  tensor<type> ret;
  FILE *fid=fopen(name,"rb");
  if (fid) {
    fseek(fid,0,SEEK_END);
    ret=tensor<type>(1,ftell(fid)/sizeof(type));
    fseek(fid,0,SEEK_SET);
    fread(ret.data,sizeof(type),ret.numel,fid);
    fclose(fid);
#ifndef NDEBUG
    printf("> %s\t<%i>\n",name,ret.numel);
  } else {
    printf("> %s\terror: not found\n",name);
#endif
  }
  return ret;
}

template<class type>
type& tensor<type>::operator()(int ind0,...) {
  int acc=1;
  va_list ap;
  va_start(ap,ind0);
  for (int c0=0;c0<ndims-1;++c0) {
    acc*=size[c0];
    ind0+=va_arg(ap,int)*acc;
  }
  va_end(ap);
  assert(0<=ind0&&ind0<numel);
  return data[ind0];
}

template<class type>
tensor<type> tensor<type>::operator()(tensor<int> ind0,...) {
  int c0,c1;
  int* sze=new int[ndims];
  tensor<int>* ind=new tensor<int>[ndims];  
  tensor<int> cum=cumsize();
  tensor<int> ins(1,ndims);
  tensor<int> ptr;
  ind[0]=ind0;
  sze[0]=ind[0].numel;
  va_list ap;
  va_start(ap,ind0);
  for (c0=1;c0<ndims;++c0) {
    ind[c0]=va_arg(ap,tensor<int>);
    sze[c0]=ind[c0].numel;
  }
  tensor<type> ret(ndims,sze);
  for (c0=0;c0<ret.numel;++c0) {
    ptr=ret.at(c0);
    for (c1=0;c1<ndims;++c1)
      ins.data[c1]=ind[c1].data[ptr.data[c1]];
    ret.data[c0]=data[ins|cum];
  }
  delete ind;
  return ret;
}

template<class type>
type& tensor<type>::last(void) {
  assert(numel);
  return data[numel-1];
} // ok

template<class type>
tensor<type> tensor<type>::operator[](int slice) {      
  tensor<type> ret(ndims-1,size);
  int c1=ret.numel*slice;
  assert(c1+ret.numel<=numel);
  for (int c0=0;c0<ret.numel;++c0)
    ret.data[c0]=data[c1++];
  return ret;      
}

template<class type>
const tensor<type>& tensor<type>::operator=(const type& data) {
  for (int c0=0;c0<numel;++c0)
    this->data[c0]=data;
  return *this;
}

template<class type>
const tensor<type>& tensor<type>::as(type data0,...) {
  va_list ap;
  va_start(ap,data0);
  data[0]=data0;
  for (int c0=1;c0<numel;++c0)
    data[c0]=va_arg(ap,type);
  va_end(ap);
  return *this;
}

template<class type>
const tensor<type>& tensor<type>::diag(type data0,...) {
  int ofs=size[0]+1;
  va_list ap;
  va_start(ap,data0);
  data[0]=data0;
  for (int c0=ofs;c0<numel;c0+=ofs)
    data[c0]=va_arg(ap,type);
  va_end(ap);
  return *this;
}

template<class type>
const tensor<type>& tensor<type>::in(type* copy) {
  for (int c0=0;c0<numel;++c0)
    data[c0]=copy[c0];
  return *this;
}

template<class type>
const tensor<type>& tensor<type>::in(int ind,type* copy,int len) {
  assert(ind+len<=numel);
  for (int c0=0;c0<len;++c0)
    data[ind+c0]=copy[c0];
  return *this;
}

template<class type>
const tensor<type>& tensor<type>::in(int ind,tensor<type> copy) {
  for (int c0=0;c0<copy.numel;++c0)
    data[ind+c0]=copy(c0);
  return *this;
}

template<class type>
template<class comp>
tensor<comp> tensor<type>::to(void) {
  tensor<comp> ret(ndims,size);
  for (int c0=0;c0<numel;++c0)
    ret.data[c0]=(comp)data[c0];
  return ret;
}

template<class type>
const tensor<type>& tensor<type>::fscan(FILE* fid) {
  char buf[48];
  for (int c0=0;c0<numel;++c0) {
    assert(!feof(fid));
    fscanf(fid,"%s",buf);
    data[c0]=(type)atof(buf);
  }
#ifndef NDEBUG
  printf("%s\n",toString());
#endif    
  return *this;
}

template<class type>
tensor<int> tensor<type>::at(int ind) {
  assert(0<=ind&&ind<numel);
  tensor<int> prd=cumsize();
  tensor<int> ret(1,ndims);
  for (int c0=ndims-1;0<=c0;--c0) {
    ret.data[c0]=ind/prd.data[c0];
    ind%=prd.data[c0];
  }
  return ret;
}

template<class type>
tensor<int> tensor<type>::cumsize(void) {
  tensor<int> ret(1,ndims);
  ret.data[0]=1;
  for (int c0=0;c0<ndims-1;++c0)
    ret.data[c0+1]=ret.data[c0]*size[c0];
  return ret;
}

template<class type>
tensor<int> tensor<type>::find(void) {
  tensor<int> ret(1,numel);
  ret.numel=0;
  for (int c0=0;c0<numel;++c0)
    if (data[c0])
      ret.data[ret.numel++]=c0;
  ret.size[0]=ret.numel;
  return ret;
}

template<class type> tensor<type> tensor<type>::diag(void) {
  assert(ndims==2);
  assert(size[0]==size[1]);
  tensor<type> ret(1,size[0]);
  for (int c0=0;c0<size[0];++c0)
    ret(c0)=(*this)(c0,c0);
  return ret;
}

template<class type> tensor<type> tensor<type>::operator*(const tensor<type>& myTensor) {
  int c0,c1,c2;
  int ofs0,ofs1;
  int tmp0,tmp1;
  int* myInt=new int[ndims+myTensor.ndims-2];
  ofs0=ofs1=1;
  for (c0=0;c0<ndims-1;++c0) {
    myInt[c0]=size[c0];
    ofs0*=size[c0];
  }
  for (c0=1;c0<myTensor.ndims;++c0) {
    myInt[ndims-2+c0]=myTensor.size[c0];
    ofs1*=myTensor.size[c0];
  }
  tensor<type> ret(ndims+myTensor.ndims-2,myInt);
  delete myInt;
  myInt=new int[myTensor.size[0]+1];
  myInt[0]=0;
  for (c0=1;c0<myTensor.size[0];++c0)
    myInt[c0]=myInt[c0-1]+ofs0;
  tmp0=tmp1=0;  
  for (c0=0;c0<ofs1;++c0) {
    for (c1=0;c1<ofs0;++c1) {
      for (c2=0;c2<myTensor.size[0];++c2)
        ret.data[tmp0]+=data[c1+myInt[c2]]*myTensor.data[tmp1+c2];
      ++tmp0;
    }
    tmp1+=myTensor.size[0];
  }
  delete myInt;
  return ret;
}

template<class type> tensor<type> tensor<type>::operator/(const tensor<type>& myTensor) {
  int ofs=0;
  tensor<int> sze(1,ndims+myTensor.ndims);
  sze.in(    0,         size,         ndims);
  sze.in(ndims,myTensor.size,myTensor.ndims);
  tensor<type> ret(sze.numel,sze.data);
  for (int c0=0;c0<myTensor.numel;++c0) {
    for (int c1=0;c1<numel;++c1)
      ret.data[c1+ofs]=data[c1]*myTensor.data[c0];    
    ofs+=numel;
  }
  return ret;
}

template<class type> tensor<type> tensor<type>::operator%(const tensor<type>& myTensor) {
  tensor<type> ret(ndims,size);
  assert(numel==myTensor.numel);
  for (int c0=0;c0<numel;++c0)
    ret.data[c0]=data[c0]*myTensor.data[c0];
  return ret;
}

template<class type> tensor<type> tensor<type>::operator+(const tensor<type>& myTensor) {
  tensor<type> ret(*this);  
  assert(numel==myTensor.numel);
  for (int c0=0;c0<numel;++c0)
    ret.data[c0]+=myTensor.data[c0];
  return ret;
}

template<class type> tensor<type> tensor<type>::operator-(const tensor<type>& myTensor) {
  tensor<type> ret(*this);
  assert(numel==myTensor.numel);
  for (int c0=0;c0<numel;++c0)
    ret.data[c0]-=myTensor.data[c0];
  return ret;
}

template<class type> tensor<type> tensor<type>::operator&(const tensor<type>& myTensor) {
  int app=myTensor.size[ndims-1];
  for (int c0=0;c0<ndims-1;++c0)
  { assert(size[c0]==myTensor.size[c0]); }
  size[ndims-1]+=app;
  tensor<type> ret(ndims,size);
  size[ndims-1]-=app;
  ret.in(    0,         data,         numel);
  ret.in(numel,myTensor.data,myTensor.numel);
  return ret;
}

template<class type> tensor<type> tensor<type>::operator*(type data) {
  tensor<type> ret(*this);
  return ret*=data;
}

template<class type> tensor<type> tensor<type>::operator/(type data) {
  tensor<type> ret(*this);
  return ret/=data; // meaningful for integer
}

template<class type> tensor<type> tensor<type>::operator+(type data) {
  tensor<type> ret(*this);
  return ret+=data;
}

template<class type> tensor<type> tensor<type>::operator-(type data) {
  tensor<type> ret(*this);
  return ret-=data;
}

template<class type> tensor<int> tensor<type>::operator<(type comp) {
  tensor<int> ret(ndims,size);
  for (int c0=0;c0<numel;++c0)
    ret.data[c0]=data[c0]<comp;
  return ret;
}

template<class type> tensor<int> tensor<type>::operator>(type comp) {
  tensor<int> ret(ndims,size);
  for (int c0=0;c0<numel;++c0)
    ret.data[c0]=data[c0]>comp;
  return ret;
}

//template<class type> tensor<int> tensor<type>::operator==(type comp) {
//  tensor<int> ret(ndims,size);
//  for (int c0=0;c0<numel;++c0)
//    ret.data[c0]=data[c0]==comp;
//  return ret;
//}
//

template<class type> int tensor<type>::operator==(const tensor<type>& myTensor) {
  int c0;
  int ret=1;
  assert(ndims==myTensor.ndims);
  for (c0=0;c0<ndims;++c0)
    assert(size[c0]==myTensor.size[c0]);
  for (c0=0;c0<numel;++c0)
    ret&=abs(data[c0]-myTensor.data[c0])<=precision; // '<=' important for type=int
  return ret;
}

template<class type> tensor<type> tensor<type>::operator&(type app) {
  assert(ndims==1);
  tensor<type> ret(1,numel+1);
  ret.in(0,data,numel);
  ret.data[numel]=app; 
  return ret;
}

template<class type> type tensor<type>::operator|(const tensor<type>& myTensor) {
  assert(numel==myTensor.numel);
  type sum=0;
  for (int c0=0;c0<numel;++c0)
    sum+=data[c0]*myTensor.data[c0];
  return sum;
}

template<class type> tensor<type> tensor<type>::x(tensor<type>& myTensor) {
  assert(numel==3||numel==4);
  assert(myTensor.numel==3||myTensor.numel==4);
  assert(ndims==1);
  assert(myTensor.ndims==1);
  assert(numel==myTensor.numel);
  tensor<type> ret(1,numel);
  ret.data[0]=data[1]*myTensor.data[2]-data[2]*myTensor.data[1];
  ret.data[1]=data[2]*myTensor.data[0]-data[0]*myTensor.data[2];
  ret.data[2]=data[0]*myTensor.data[1]-data[1]*myTensor.data[0];
  return ret;
}

template<class type> const tensor<type>& tensor<type>::operator+=(const tensor<type>& myTensor) {
  assert(ndims==myTensor.ndims);
  assert(numel==myTensor.numel);
  for (int c0=0;c0<numel;++c0)
    data[c0]+=myTensor.data[c0];
  return *this;
}

template<class type> const tensor<type>& tensor<type>::operator-=(const tensor<type>& myTensor) {
  assert(numel==myTensor.numel);
  for (int c0=0;c0<numel;++c0)
    data[c0]-=myTensor.data[c0];
  return *this;
}

template<class type> const tensor<type>& tensor<type>::operator*=(type data) {
  for (int c0=0;c0<numel;++c0)
    this->data[c0]*=data;
  return *this;
}

template<class type> const tensor<type>& tensor<type>::operator/=(type data) {
  for (int c0=0;c0<numel;++c0)
    this->data[c0]/=data;
  return *this;
}

template<class type> const tensor<type>& tensor<type>::operator+=(type data) {
  for (int c0=0;c0<numel;++c0)
    this->data[c0]+=data;
  return *this;
}

template<class type> const tensor<type>& tensor<type>::operator-=(type data) {
  for (int c0=0;c0<numel;++c0)
    this->data[c0]-=data;
  return *this;
}

template<class type> char* tensor<type>::toInfo(void) {
  int len=sprintf(TENSOR_BUFFER,"tensor<%i>([",sizeof(type));
  for (int c0=0;c0<ndims;c0++)
    len+=sprintf(TENSOR_BUFFER+len,"%ix",size[c0]);
  len-=ndims>0;
  len+=sprintf(TENSOR_BUFFER+len,"],%i,",numel);
  if (!data)
    len+=sprintf(TENSOR_BUFFER+len,"null)");
  else
    len+=sprintf(TENSOR_BUFFER+len,"%08x)",data);
  return TENSOR_BUFFER;
}

template<class type> char* tensor<type>::toString(char* str,char* fmt) {
  int c0,c1,c2,c3;
  char tmp[16];
  // define expression for 0 value
  char nul[16];
  c1=0;
  sprintf(nul,fmt,(type)0);
  for (c0=1;c0<strlen(nul);++c0)
    if (nul[c0]=='.')
      c1=c0;
  if (c1) {
    for (c0=0;c0<strlen(nul);++c0)
      nul[c0]=' ';
    nul[c1-1]='0';
    nul[c1  ]='.';
  }
  // format tensor
  str[0]=0;
  int len=0;
  if (fmt&&data&&numel)
    switch (ndims) {
      case 0:
        TENSOR_FORMAT(data[c0])
        break;
      case 1:
        for (c0=0;c0<size[0];c0++)
          TENSOR_FORMAT(data[c0])
        break;
      case 2:
        for (c0=0;c0<size[0];++c0) {
          for (c1=0;c1<size[1];++c1)
            TENSOR_FORMAT((*this)(c0,c1))
          if (c0+1<size[0])
            len+=sprintf(str+len,"\n");
        }
        break;
      case 3:
        tmp[0]=0;
        if (size[0]<2)
          strcat(tmp,"{:,:,%i}"); // maybe use strcat please
        else
          strcat(tmp,"{:,:,%i}\n");
        for (c2=0;c2<size[2];++c2) {
          len+=sprintf(str+len,tmp,c2);
          for (c0=0;c0<size[0];++c0) {
            for (c1=0;c1<size[1];++c1)
              TENSOR_FORMAT((*this)(c0,c1,c2))
            len+=sprintf(str+len,"\n");
          }
        }
        break;
      case 4:
        tmp[0]=0;
        if (size[0]<2)
          strcat(tmp,"{:,:,%i,%i}");
        else
          strcat(tmp,"{:,:,%i,%i}\n");
        for (c3=0;c3<size[3];++c3)
        for (c2=0;c2<size[2];++c2) {
          len+=sprintf(str+len,tmp,c2,c3);
          for (c0=0;c0<size[0];++c0) {
            for (c1=0;c1<size[1];++c1)
              TENSOR_FORMAT((*this)(c0,c1,c2,c3))
            len+=sprintf(str+len,"\n");
          }
        }
        break;
      default:
        assert(0);
    }
  return str;
}

template<class type>
char* tensor<type>::toString(char* fmt) {
  return toString(TENSOR_BUFFER,fmt);
}

template<class type>
char* tensor<type>::toString(void) {
  return toString(tensor<type>::format);
}

template<class type> tensor<type> tensor<type>::exp(void) {
  assert(ndims==2&&size[0]&&size[1]);
  tensor<type> uni=*this;
  tensor<type> sum=eye(size[0]);
  tensor<type> nxt=eye(size[0]);
  for (int c0=1;c0<100&&TENSOR_EPS<nxt.norm(2);++c0) {
    nxt=nxt*uni;
    nxt*=1/(type)c0;
    sum+=nxt;
    //printf("%3i\t%10.2e\n",c0,nxt.norm(2));
  }
  return sum;
}

template<class type> tensor<type> tensor<type>::log(void) {
  assert(ndims==2&&size[0]&&size[1]);
  tensor<type> uni=*this-eye(size[0]);
  tensor<type> sum=uni;
  tensor<type> nxt=uni;
  for (int c0=2;c0<100&&TENSOR_EPS<nxt.norm(2);++c0) {
    nxt=nxt*uni;
    nxt*=(1-c0)/(type)c0;
    sum+=nxt;
    //printf("%3i\t%10.2e\n",c0,nxt.norm(2));
  }
  return sum;
}

template<class type> tensor<type> tensor<type>::transpose(tensor<int>& sig) {
  assert(ndims==sig.numel);
  int c0,c1;
  tensor<type> ret(ndims,size);
  for (c0=0;c0<ndims;++c0)
    ret.size[sig(c0)]=size[c0];
  tensor<int> cum=ret.cumsize();
  tensor<int> loc;
  tensor<int> ins(1,ndims);
  for (c0=0;c0<numel;++c0) {
    loc=at(c0);
    for (c1=0;c1<ndims;++c1)
      ins.data[sig(c1)]=loc(c1);
    ret.data[ins|cum]=data[c0];
  }
  return ret;
}

template<class type> tensor<type> tensor<type>::transpose(void) {
  tensor<int> sig=tensor<int>::vector(2).as(1,0);
  return transpose(sig);
}

template<class type> tensor<type> tensor<type>::trace(tensor<int>& dims) {
  int c0,c1,c2;
  for (c0=0;c0<dims.numel;++c0) {
    assert(dims(c0)<ndims);
  }
  int sze0=size[dims(0)];
  for (c0=1;c0<dims.numel;++c0) {
    assert(dims(c0-1)<dims(c0));
    assert(sze0==size[dims(c0)]);
  }
  tensor<int> cum=cumsize();
  tensor<int> sig(1,ndims-dims.numel);
  c1=0;
  for (c0=0;c0<sig.numel;++c0) {
    while ((dims==c1).find().numel)
      c1++;
    sig.data[c0]=c1++;
  }
  tensor<int> sze(1,ndims-dims.numel);
  for (c0=0;c0<sig.numel;++c0)
    sze.data[c0]=size[sig.data[c0]];
  tensor<type> ret(sze.numel,sze.data);
  tensor<int> ind(1,ndims);
  tensor<int> atr;
  for (c0=0;c0<ret.numel;++c0) {
    atr=ret.at(c0);
    for (c1=0;c1<sig.numel;++c1)
      ind.data[sig.data[c1]]=atr(c1);
    for (c1=0;c1<sze0;++c1) {
      for (c2=0;c2<dims.numel;++c2)
        ind.data[dims(c2)]=c1;      
      ret.data[c0]+=data[ind|cum]; 
    }   
  }
  return ret;
}

template<class type> tensor<type> tensor<type>::trace(int dim0,int dim1) {
  tensor<int> dims=tensor<int>::vector(2).as(dim0,dim1);
  return trace(dims);
}

template<class type> type tensor<type>::norm(int norm) {
  type ret=0;
  type cmp;
  switch (norm) {
    case 0: // infinity norm
      for (int c0=0;c0<numel;++c0) {
        cmp=abs(data[c0]);
        ret=(ret<cmp)?cmp:ret;
      }
      return ret;
    case 1: // sum of absolutes
      for (int c0=0;c0<numel;++c0)
        ret+=abs(data[c0]);
      return ret;
    case 2: // Euclidean norm
      for (int c0=0;c0<numel;++c0)
        ret+=data[c0]*data[c0];
      return sqrt(ret);
  }
  return 0;
}

template<class type>
tensor<type> tensor<type>::normal(int norm) {
  type sca=this->norm(norm);
  //assert(TENSOR_EPS<sca);
  if (valid=TENSOR_NO_DIVISION<sca)  
    return *this*(1/sca);
  else
    return *this;
}

template<class type>
inline tensor<type> tensor<type>::normal() {
  return normal(2);
}

template<class type> tensor<type> tensor<type>::absolute(void) {
  tensor<type> ret(*this);
  for (int c0=0;c0<numel;++c0)
    ret.data[c0]=abs(data[c0]);
  return ret;
}

template<class type> tensor<type> tensor<type>::sort(tensor<int>& ind) {  
  ind=tensor<int>::index(1,numel);
  tensor<type> ret(*this);
  ret.quicksort(0,numel-1,ind.data);
  return ret;
}

template<class type> void tensor<type>::quicksort(int a,int b,int *ind) {
  int c1=a,c2=b;
  int ind_;
  type ref=data[a+b>>1]; // selects middle index as reference
  type dat_;
  do {
    while (data[c1]<ref) ++c1;
    while (data[c2]>ref) --c2;
    if (c1<=c2) {
          ind_=ind[c1];
       ind[c1]=ind[c2];
       ind[c2]=ind_;
          dat_=data[c1];
      data[c1]=data[c2];
      data[c2]=dat_;
      ++c1;
      --c2;
    }
  } while (c1<=c2);
  if (a<c2  ) quicksort(a,c2,ind);
  if (  c1<b) quicksort(c1,b,ind);
}

template<class type> tensor<type> tensor<type>::tros(tensor<int>& ind) {  
  ind=tensor<int>::index(1,numel);
  tensor<type> ret(*this);
  ret.quicktros(0,numel-1,ind.data);
  return ret;
}

template<class type> void tensor<type>::quicktros(int a,int b,int *ind) {
  int c1=a,c2=b;
  int ind_;
  type ref=data[a+b>>1]; // selects middle index as reference
  type dat_;
  do {
    while (data[c1]>ref) ++c1;
    while (data[c2]<ref) --c2;
    if (c1<=c2) {
          ind_=ind[c1];
       ind[c1]=ind[c2];
       ind[c2]=ind_;
          dat_=data[c1];
      data[c1]=data[c2];
      data[c2]=dat_;
      ++c1;
      --c2;
    }
  } while (c1<=c2);
  if (a<c2  ) quicktros(a,c2,ind);
  if (  c1<b) quicktros(c1,b,ind);
}

template<class type> tensor<type> tensor<type>::apply(type (*func)(type)) {
  tensor<type> ret(*this);
  for (int c0=0;c0<numel;++c0)
    ret.data[c0]=func(data[c0]);
  return ret;
}

template<class type> const tensor<type>& tensor<type>::apply(type (*func)(void)) {
  for (int c0=0;c0<numel;++c0)
    data[c0]=func();
  return *this;
}

template<class type> type tensor<type>::sum(void) {
  type ret=0;
  for (int c0=0;c0<numel;++c0)
    ret+=data[c0];
  return ret;
}

template<class type> type tensor<type>::prod(void) {
  type ret=1;
  for (int c0=0;c0<numel;++c0)
    ret*=data[c0];
  return ret;
}

template<class type> type tensor<type>::min_(int& ind) {
  type ret=data[ind=0];
  for (int c0=1;c0<numel;++c0)
    if (data[c0]<ret)
      ret=data[ind=c0];        
  return ret;
}

template<class type> type tensor<type>::max_(int& ind) {
  type ret=data[ind=0];
  for (int c0=1;c0<numel;++c0)
    if (data[c0]>ret)
      ret=data[ind=c0];        
  return ret;
}

template<class type> tensor<type> tensor<type>::chop(void) {
  tensor<type> ret(ndims,size);
  for (int c0=0;c0<numel;++c0)
    if (TENSOR_NO_DIVISION<fabs(data[c0]))
      ret.data[c0]=data[c0];
  return ret;
}

template<class type>
tensor<tensor<type>*> tensor<type>::split(type sep) { // only one separator possible (because can be discarded) 
  int c0,c1;
  int cnt=1;
  for (c0=0;c0<numel;++c0)
    if (data[c0]==sep)
      cnt++;
  tensor<tensor<type>*> ret(1,cnt);
  c1=cnt=0;
  for (c0=0;c0<numel;++c0)
    if (data[c0]==sep) {
      //printf("%i\t%i %i\n",cnt,c1,c0);          
      ret(cnt)=new tensor<type>(1,c0-c1);
      ret(cnt)->in(0,&data[c1],c0-c1);
      //ret(cnt)->in(c0-c1,app.data,app.numel);
      cnt++;
      c1=c0+1;
    }
  //printf("%i\t%i %i\n",cnt,c1,c0);          
  ret(cnt)=new tensor<type>(1,c0-c1);
  ret(cnt)->in(0,&data[c1],c0-c1);
  //ret(cnt)->in(c0-c1,app.data,app.numel);
  return ret;
}

// format: ndims,size[0:ndims-1],data
template<class type>
void tensor<type>::store(char* name) {
  FILE *fid=fopen(name,"wb");
  if (fid) {
    fwrite(&ndims,sizeof(int),1,fid);
    fwrite(size,sizeof(int),ndims,fid);
    fwrite(data,sizeof(type),numel,fid);
    fclose(fid);
#ifndef NDEBUG
    printf("%s <-\t%s\n",name,toInfo());
  } else {
    printf("%s <- (error)\t%s\n",name,toInfo());
#endif
  }
}

template<class type>
void tensor<type>::flush(char* name) {
  FILE *fid=fopen(name,"wb");
  if (fid) {
    fwrite(data,sizeof(type),numel,fid);
    fclose(fid);
#ifndef NDEBUG
    printf("%s <-\t%s\n",name,toInfo());
  } else {
    printf("%s <- (error)\t%s\n",name,toInfo());
#endif
  }
}

template<class type>
void tensor<type>::free(void) {
  for (int c0=0;c0<numel;++c0) {
    assert(data[c0]);
    delete data[c0];
  }
}

#include "tensor_math.hpp"

char* tensor<char>::format="%c";
char* tensor<int>::format="%i ";
char* tensor<float>::format="%10.2e ";
char* tensor<double>::format="%10.2e ";
char* tensor<void*>::format="%08x ";
int tensor<int>::valid=1;
int tensor<float>::valid=1;
int tensor<double>::valid=1;
int tensor<int>::precision=0;
float tensor<float>::precision=1e-5;
double tensor<double>::precision=1e-10;

#endif
