// code by jph

// svd() from 'Numerical Recipes in C++' by Press, Teukolsky, Vetterling, Flannery

#define TENSOR_EIG          1E-10
#define TENSOR_QR           1E-06
#define TENSOR_SCHUR        1E-06
#define TENSOR_SVD          30
#define TENSOR_RANK         1E-07

// TODO, normal() decides about the numerics in qr - is that good ?
template<class type>
tensor<type> tensor<type>::solve(tensor<type> rhs) {
  assert(ndims==2);
  assert(size[0]==size[1]);
  assert(size[1]==rhs.size[0]);
  int c0,c1,c2,c1s=0,k,swap;
  type tfac,tmax,tabs;
  type* copy=new type[numel];
  for (c0=0;c0<numel;++c0)
    copy[c0]=data[c0];
  tensor<type> sol(2,rhs.size[0],(rhs.ndims<=1)?1:rhs.size[1]);
  tensor<int> ind=tensor<int>::index(1,size[0]);
  valid=1;
  for (c0=0;c0<size[0];++c0) {
    k=c0;
    tmax=0;
    for (c1=c0;c1<size[0];++c1)
      if (tmax<(tabs=abs(copy[ind.data[c1]+c1s]))) {
        tmax=tabs;
        k=c1;
      }
    swap=ind.data[k];
    ind.data[k]=ind.data[c0];
    ind.data[c0]=swap;
    for (c1=c0+1;c1<size[0];++c1) {
      assert(TENSOR_NO_DIVISION<abs(copy[ind.data[c0]+c1s]));
      if (abs(copy[ind.data[c0]+c1s])<TENSOR_NO_DIVISION) {
        //printf("fail type 1\n");
        valid=0; // TODO bring inti if
        break;
      }
      tabs=copy[ind.data[c1]+c1s]/copy[ind.data[c0]+c1s]; // division
      for (c2=(c0+1)*size[0];c2<numel;c2+=size[0])
        copy[ind.data[c1]+c2]-=copy[ind.data[c0]+c2]*tabs;
      for (c2=0;c2<rhs.numel;c2+=size[0])
        rhs.data[ind.data[c1]+c2]-=rhs.data[ind.data[c0]+c2]*tabs;
    }
    c1s+=size[0];
  }
  //printf("%s\n",copy.toString());
  //printf("%s\n",ind.toString());
  for (c0=size[0]-1;c0>=0;--c0) {
    assert(TENSOR_NO_DIVISION<abs(copy[ind.data[c0]+c0*size[0]]));
    if (abs(copy[ind.data[c0]+c0*size[0]])<TENSOR_NO_DIVISION) {
      //printf("fail type 2\n");
      valid=0;
      break;
    }
    tfac=1/copy[ind.data[c0]+c0*size[0]]; // division
    for (c1=0;c1<sol.size[1];++c1) {
      tabs=rhs.data[ind.data[c0]+c1*size[0]];
      for (c2=c0+1;c2<size[0];++c2)
        tabs-=copy[ind.data[c0]+c2*size[0]]*sol.data[c2+c1*size[0]];
      sol.data[c0+c1*size[0]]=tabs*tfac;
    }
  }
  sol.ndims=rhs.ndims;
  delete[] copy;
  return sol;
}



template<class type>
tensor<type> tensor<type>::qr(tensor<type>& ret) {
  ret=*this;
  tensor<type> qor=eye(size[0]);
  tensor<type> col(2,size[0],1);
  tensor<type> row(2,1,size[0]);
  tensor<type> tmp;
  for (int c0=0;c0<min(size[0]-1,size[1]);++c0) {
    tensor<type> prj(1,size[0]);
    prj.in(c0,&ret.data[c0+c0*size[0]],size[0]-c0);
    prj.data[c0]-=prj.norm(2);
    prj=prj.normal(2);
    col.in(prj.data);
    row.in(prj.data);
    col*=2;
    tmp=eye(size[0])-col*row;
    ret=tmp*ret;
    qor=tmp*qor;
  }
  assert(qor.transpose()==qor.solve(tensor<double>::eye(qor.size[0]))); // qor is orthogonal
  assert(ret==qor**this);                                               // R=Q*A <=> Q^t*R=A
  for (int c1=0;c1<size[1];++c1)
    for (int c0=c1+1;c0<size[0];++c0)
      assert(abs(ret(c0,c1))<TENSOR_QR); // values below diagonal are null
  return qor;
}

template<class type>
tensor<type> tensor<type>::schur(tensor<type>& ret) { // sor*ret*sor^t=*this
  assert(ndims==2);
  assert(size[0]==size[1]);
  tensor<double> sor=eye(size[0]);
  tensor<double> qor;
  tensor<double> tmp;
  tensor<double> cpy;
  ret=*this;
  for (int c0=0;c0<100;++c0) {
    cpy=ret.diag();
    qor=ret.qr(tmp);
    ret=qor*ret*qor.transpose();
    sor=qor*sor;
    valid=1;
    for (int c1=0;c1<size[1];++c1)
      for (int c2=c1+1;c2<size[0];++c2)
        valid&=abs(ret(c2,c1))<TENSOR_SCHUR; // values below diagonal are null
    printf("%i %s\n---\n",c0,ret.toString());
    if (valid)
      break;
  }
  //valid=c0<100;
  return sor.transpose();
}

// right now:   ret * diag(w) * v' = *this
template<class type>
tensor<type> tensor<type>::svd(tensor<type>& w,tensor<type>& v) {
  assert(ndims==2);
  assert(size[1]<=size[0]);
  int flag;
  int i,its,j,jj,k,l,nm;
  type anorm,c,f,g,h,s,scale,x,y,z;
  int m=size[0];
  int n=size[1];
  tensor<type> ret(*this); // U = A
  tensor<type> rv1(1,n);
  w=tensor<type>(1,n);     // w is output vector as diagonal of W
  v=tensor<type>(2,n,n);   // v is output matrix as V
  valid=1;
  // householder reduction to bidiagonal form
  g=scale=anorm=0;
  for (i=0;i<n;i++) {
    l=i+2;
    rv1(i)=scale*g;
    g=s=scale=0;
    if (i<m) {
      for (k=i;k<m;k++)
        scale+=abs(ret(k,i));
      if (TENSOR_NO_DIVISION<abs(scale)) {
        for (k=i;k<m;k++) {
          ret(k,i)/=scale;
          s+=ret(k,i)*ret(k,i);
        }
        f=ret(i,i);
        g=-math::Sign(sqrt(s),f);
        h=f*g-s;
        ret(i,i)=f-g;
        for (j=l-1;j<n;j++) {
          for (s=0,k=i;k<m;k++)
            s+=ret(k,i)*ret(k,j);
          f=s/h;
          for (k=i;k<m;k++)
            ret(k,j)+=f*ret(k,i);
        }
        for (k=i;k<m;k++)
          ret(k,i)*=scale;
      }
    }
    w(i)=scale*g;    
    g=s=scale=0;
    if (i+1<=m&&i!=n) {
      for (k=l-1;k<n;k++)
        scale+=abs(ret(i,k));
      if (TENSOR_NO_DIVISION<abs(scale)) {
        for (k=l-1;k<n;k++) {
          ret(i,k)/=scale;
          s+=ret(i,k)*ret(i,k);          
        }
        f=ret(i,l-1);
        g=-math::Sign(sqrt(s),f);
        h=f*g-s;
        ret(i,l-1)=f-g;        
        for (k=l-1;k<n;k++)
          rv1(k)=ret(i,k)/h;        
        for (j=l-1;j<m;j++) {
          for (s=0,k=l-1;k<n;k++)
            s+=ret(j,k)*ret(i,k);          
          for (k=l-1;k<n;k++)
            ret(j,k)+=s*rv1(k);
        }
        for (k=l-1;k<n;k++)
          ret(i,k)*=scale;
      }
    }
    anorm=max(anorm,(abs(w(i))+abs(rv1(i))));
  }
  // accumulation of right hand transformations
  for (i=n-1;i>=0;i--) {
    if (i<n-1) {
      if (TENSOR_NO_DIVISION<abs(g)) {
        for (j=l;j<n;j++)
          v(j,i)=ret(i,j)/ret(i,l)/g;
        for (j=l;j<n;j++) {
          for (s=0,k=l;k<n;k++)
            s+=ret(i,k)*v(k,j);
          for (k=l;k<n;k++)
            v(k,j)+=s*v(k,i);
        }
      }
      for (j=l;j<n;j++)
        v(i,j)=v(j,i)=0;
    }
    v(i,i)=1;
    g=rv1(i);
    l=i;
  }
  for (i=min(m,n)-1;i>=0;i--) {
    l=i+1;
    g=w(i);
    for (j=l;j<n;j++)
      ret(i,j)=0;
    if (TENSOR_NO_DIVISION<abs(g)) {
      g=1/g;
      for (j=l;j<n;j++) {
        for (s=0,k=l;k<m;k++)
          s+=ret(k,i)*ret(k,j);
        f=s/ret(i,i)*g;
        for (k=i;k<m;k++)
          ret(k,j)+=f*ret(k,i);
      }
      for (j=i;j<m;j++)
        ret(j,i)*=g;
    } else
      for (j=i;j<m;j++)
        ret(j,i)=0;
    ret(i,i)+=1;
  }
  // diagonalization of the bidiagonal form
  for (k=n-1;k>=0;k--) {
    for (its=0;its<=TENSOR_SVD;its++) {
      flag=1;
      for (l=k;l>=0;l--) {
        nm=l-1;
        if (abs(rv1(l))<TENSOR_NO_DIVISION) {
          flag=0;
          break;
        }
        if (abs(w(nm))<TENSOR_NO_DIVISION)
          break;        
      }
      if (flag) {
        c=0;
        s=1;
        for (i=l-1;i<k+1;i++) {
          f=s*rv1(i);
          rv1(i)=c*rv1(i);
          if (abs(f)<TENSOR_NO_DIVISION)
            break;
          g=w(i);
          h=sqrt(f*f+g*g);
          w(i)=h;
          h=1/h;
          c=g*h;
          s=-f*h;
          for (j=0;j<m;j++) {
            y=ret(j,nm);
            z=ret(j,i);
            ret(j,nm)=y*c+z*s;
            ret(j,i)=z*c-y*s;            
          }
        }
      }
      z=w(k);
      if (l==k) {
        if (z<0) {
          w(k)=-z;
          for (j=0;j<n;j++)
            v(j,k)=-v(j,k);
        }
        break;
      }
      valid&=its<TENSOR_SVD;
      x=w(l);
      nm=k-1;
      y=w(nm);
      g=rv1(nm);
      h=rv1(k);
      f=((y-z)*(y+z)+(g-h)*(g+h))/(2*h*y);
      g=sqrt(f*f+1);
      f=((x-z)*(x+z)+h*((y/(f+math::Sign(g,f)))-h))/x;
      c=s=1;
      for (j=l;j<=nm;j++) {
        i=j+1;
        g=rv1(i);
        y=w(i);
        h=s*g;
        g=c*g;
        z=sqrt(f*f+h*h);
        rv1(j)=z;
        c=f/z;
        s=h/z;
        f=x*c+g*s;
        g=g*c-x*s;
        h=y*s;
        y*=c;
        for (jj=0;jj<n;jj++) {
          x=v(jj,j);
          z=v(jj,i);
          v(jj,j)=x*c+z*s;
          v(jj,i)=z*c-x*s;
        }
        z=sqrt(f*f+h*h);
        w(j)=z;
        if (TENSOR_NO_DIVISION<abs(z)) {
          z=1/z;
          c=f*z;
          s=h*z;
        }
        f=c*g+s*y;
        x=c*y-s*g;
        for (jj=0;jj<m;jj++) {
          y=ret(jj,j);
          z=ret(jj,i);
          ret(jj,j)=y*c+z*s;
          ret(jj,i)=z*c-y*s;
        }
      }
      rv1(l)=0;
      rv1(k)=f;
      w(k)=x;
    }
  }
  tensor<int> ind;
  w=w.tros(ind);
  tensor<type> prm(2,n,n);
  for (i=0;i<n;i++)
    prm(ind(i),i)=1;
  v=v*prm; // TODO optimize
  return (ret*prm).transpose(); // TODO optimize
}

// if this=U*D*V'  then  pinv(this)=V*D^(-1)*U'
template<class type>
tensor<type> tensor<type>::inverse(void) {
  int cnt=0;
  type fac;
  tensor<type> val,v;
  tensor<type> ut=svd(val,v);
  valid=0;
  printf("%s \n",val.toString());
  for (int c1=0;c1<val.numel;++c1)
    if (precision<val.data[c1]) {
      ++valid;
      fac=1/val.data[c1];
      for (int c0=0;c0<v.size[0];c0++)
        v.data[cnt++]*=fac;
    } else
      break;
  for (int c1=cnt;c1<v.numel;++c1)
    v.data[c1]=0;
  //printf("final V\n%s\n-----------\n",ort.toString());
  return v*ut;
}

template<class type>
tensor<type> tensor<type>::inverse(int num) {
  int cnt=0;
  type fac;
  tensor<type> val,v;
  tensor<type> ut=svd(val,v);
  //assert(ut.transpose()*tensor<double>::diag(val)*v.transpose()==*this); // fails !!! svd alg mmmh!!!
  valid=0;
  for (int c1=0;c1<num;++c1)
    if (precision<val.data[c1]) {
      ++valid;
      fac=1/val.data[c1];
      for (int c0=0;c0<v.size[0];c0++)
        v.data[cnt++]*=fac;
    } else {
      printf("%s\n",(ut.transpose()*tensor<double>::diag(val)*v.transpose()-*this).toString());
      printf("warning: rank smaller than expected!\n");
      printf("values are: %s\n",val.toString());
      break;
    }
  for (int c1=cnt;c1<v.numel;++c1)
    v.data[c1]=0;
  return v*ut;
}

// if this=U*D*V'  then  pinv(this)=V*D^(-1)*U'
template<class type>
tensor<type> tensor<type>::null(void) {
  //tensor<type> dia,ort;
  //tensor<type> urt=svd(dia,ort);
  //for (int c1=0;c1<dia.numel;++c1)
  //  if (TENSOR_NO_DIVISION<dia.data[c1])
  //    dia.data[c1]=1/dia.data[c1];
  return *this;
}

template<class type>
int tensor<type>::rank(void) {
  int c0;
  tensor<type> val,ort;
  if (size[0]<size[1])
    transpose().svd(val,ort);
  else
    svd(val,ort);
  for (c0=0;c0<val.numel;c0++)
    if (val.data[c0]<=precision)
      break;
  return c0;
}
