// code by jph

#define ANI_HUMANIZED   0x00000001
#define ANI_DRAW_DYN    0x00000002
// generic  band [1,2,4,8,16,32,64,128]
// specific band [256,]

class animated {
  public:
		animated(void);
		virtual ~animated(void);
		virtual void control(void) {};
		virtual void motion(void); // rarely overloaded!
		virtual void draw(void)=0;
		void draw_std(void);
		virtual void draw_dyn(void);
		void setEllipsoid(const tensor<double>& len);
    tensor<double> Iine(void); // 4x4 inverse inertia
    tensor<double> inertia(tensor<double> wrt); // 4x4 inertia with respect to wrt
    virtual tensor<double> momentum_lin(void); // linear momentum
    virtual tensor<double> momentum_ang(void); // angular momentum
    double energy_kin(void);
    tensor<double> myF(tensor<double> arg);
    tensor<double> link_lhs(tensor<double> vec,tensor<double> lev);
    tensor<double> link_rhs(tensor<double> vec);
    tensor<double> link_ect(tensor<double> vec);
//----------------------------------------------------------------------
		tensor<double> pos; // 4x4 state tensor (orientation, position)
		tensor<double> vel; // 4   velocity WORLD
		tensor<double> ome; // 4   angular velocity OBJECT
		tensor<double> acc; // 4   accel OBJECT coordinates
		tensor<double> tor; // 4   torque tensor 
		tensor<double> fva;
		tensor<double> fot;
		tensor<double> ine; // 4x4 inertia tensor
		tensor<double> col; // 4   color
		tensor<double> ext; // 4   extremities object/initialized by setEllipsoid (used for intertia/mass computation) 
		double mass;
		time rmv;
		unsigned int status;
//-------------------------------------------------------------------------------------------------		
    static int motioned;
    static double time;
    static double dt;
};

animated::animated(void) {
  pos=tensor<double>::eye(4);
  vel=tensor<double>::vector(4);
  ome=tensor<double>::vector(4);
  acc=tensor<double>::vector(4);
  tor=tensor<double>::vector(4);
  fva=tensor<double>::vector(4);
  fot=tensor<double>::vector(4);
  ine=tensor<double>::matrix(4,4);
  col=tensor<double>::vector(4).as( 1.0, 1.0, 1.0, 1.0);  
  setEllipsoid(tensor<double>::vector(4).as( 1.0, 1.0, 1.0, 0.0));
};

animated::~animated(void) {}; // no news

void animated::motion(void) { // input: acc (world), tor (object)
  int c1;
  // acc --> vel
  acc-=glI(pos)*vel%fva; // for avatar
  // acc*=dt;
  vel+=pos*acc*dt;
  // acc=0; // TODO later say acc=0;
  tor-=ome%fot; // for avatar
  ome+=Iine()*( tor - ome.x(ine*ome) )*dt; // tor (object)
  // tor=0; // TODO later say tor=0 !!!
  // vel --> pos
  for (int c0=0;c0<3;++c0)
    pos.data[12+c0]+=vel.data[c0]*dt;
  pos=pos*glD(ome*dt).exp();
  ++motioned;
}

void animated::draw_std(void) {
  glPushMatrix();
    glMultMatrixd(pos.data);
    glColor4dv(col.data);
    draw();
  glPopMatrix();
}

void animated::draw_dyn(void) { // right now: call from world coords
  double alp=0.7;
  glPushAttrib(GL_DEPTH_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);
    glPushMatrix();
      glColor4d( 0.8, 0.8, 0.8, alp);
      glMultMatrixd(glMatrixFromVector(momentum_lin(),pos[3]).data);
      glutSolidCylinder(0,momentum_lin().norm(2),.025,10);
    glPopMatrix();
    glPushMatrix();
      glColor4d( 1.0, 1.0, 0.0, alp);
      glMultMatrixd(glMatrixFromVector(momentum_ang(),pos[3]).data);
      glutSolidCylinder(0,momentum_ang().norm(2),.025,10);
    glPopMatrix();
  glPopAttrib();  
}

void animated::setEllipsoid(const tensor<double>& len) {    
  ext=len;
  ext(3)=1; // TODO why not make this 0 ?
  mass=4.188790205*ext(0)*ext(1)*ext(2);
  assert(TENSOR_EPS<mass);
  ine.diag( 0.2*mass*(ext(1)*ext(1)+ext(2)*ext(2)),
            0.2*mass*(ext(0)*ext(0)+ext(2)*ext(2)),
            0.2*mass*(ext(0)*ext(0)+ext(1)*ext(1)),
            0.0);
}

tensor<double> animated::Iine(void) {
  return tensor<double>::matrix(4,4).diag(1/ine.data[ 0],1/ine.data[ 5],1/ine.data[10],1.0);
}

tensor<double> animated::inertia(tensor<double> wrt) {
  tensor<double> ofs=wrt.solve(pos);
  tensor<double> rot=gl3m(ofs);
  tensor<double> lin=gl3v(ofs[3]);
  tensor<double> tmp=gl3m(ine);
  tensor<double> ret=rot.transpose()*tmp*rot+(tensor<double>::eye(3)*(lin|lin)-lin/lin)*mass;
  ret=gl4r(ret);
  ret(3,3)=0;
  return ret;
}

tensor<double> animated::momentum_lin(void) {
  return vel*mass;
}

tensor<double> animated::momentum_ang(void) {
  return pos*ine*ome+pos[3].x(vel)*mass;
}

double animated::energy_kin(void) {
  return 0.5*((vel|vel*mass)+(ine*ome|ome));
}

tensor<double> animated::link_lhs(tensor<double> vec,tensor<double> lev) {
  return gl3m(pos*glD(vec)*Iine()*glD(lev)*glI(pos));
}

tensor<double> animated::link_rhs(tensor<double> vec) {
  return gl3v(pos*glD(ome)*ome.x(vec)+pos*glD(vec)*Iine()*ome.x(ine*ome));
}

tensor<double> animated::link_ect(tensor<double> vec) {
  return gl3m(pos*glD(vec)*Iine()*glI(pos));
}

int animated::motioned=0; // temp
double animated::time=0;
double animated::dt=0;
